/*
Copyright 2007, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2007
*/

#include <ScatterWeb.System.h>
#include <ScatterWeb.CC1020.h>
#include <ScatterWeb.Cmd.internal.h>
#include <string.h>

#include "log.h"

#include "controller.basicIO.h"
#include "controller.trafficGenerator.h"
#include "controller.netTypes.h"

#include "aodv.messageTypes.h"
#include "aodv.routing.h"
#include "aodv.interface.h"
#include "aodv.tests.h"


#define LOGTAG		("application")


bool Process_radioHandler(struct netpacket_handler_args* args) {
	/*
		Notes: * The received packet data becomes invalid when this handler is completed.
	 */

	// Pass packet through command interface
	if ( Cmd_radioHandler( args ) ) {
		// packet handled by messaging module
		return true;
	}

	/*
	 * Traffic-Generator-packets will be dumped.
	 */
	if (args->netheader->type == TRAFFIC_GENERATOR_PACKET) {
		dumpPacket(args, false);
		return true;
	}

	if (args->netheader->type == AODV_TEST_PACKAGE) {
		processTestPacket(args->payload, args->payload_size);
		return true;
	}

	/*
	 * AODV-Control messages are captured here.
	 */
	if (args->netheader->type == AODV_CONTROL_MESSAGE) {
		/**
		 * Packet contains an AODV control message.
		 * Pass it to the AODV protocol.
		 */
		log("layer-2", "AODV-Control-Packet received from '%i'. Passing to AODV protocol.", args->netheader->from);
		aodvProcessControlMessage(args);
		logsep();
		return true;
	}

	/*
	 * AODV-Forward messages are captured here.
	 */
	if (args->netheader->type == AODV_DATA_PACKAGE && args->cpaResult == ADRCOMP_THIS) {
		/**
		 * Packet contains a packet that was sent to this node using the AODV protocol.
		 * This packet could be either destined towards >this< node or
		 * to another node.
		 */
		log("layer-2", "AODV-Data-Packet received from '%i'. Passing to AODV protocol.", args->netheader->from);
		aodvProcessDataPacket(args);
		logsep();
		return true;
	}

	/*
	 * OTHER Nodes
	 */
	if (args->cpaResult == ADRCOMP_OTHER) {
		return false;
	}

	/*
	 * BROADCAST packet
	 */
	else if (args->cpaResult == ADRCOMP_BROADCAST) {
		// process broadcast packets here

		switch (args->netheader->type) {
			case PING_REQUEST_PACKET:
				// send back to receiver's address
				Net_sendPong(args->netheader->from);
				break;
			default:
				return false;		// packet not handled (unknown)
		}

	}

	/*
	 * THIS node
	 */
	else {
		// process packets for this node here
		switch (args->netheader->type) {
			case PING_REQUEST_PACKET:
				// send back to receiver's address
				Net_sendPong(args->netheader->from);
				break;

			case PING_REPLY_PACKET:
				LOG_LOW("PONG from %u", args->netheader->from);
				break;

			default:
				return false;		// packet not handled (unknown)
		}
		return true;				// packet handled
	}
	return false;					// packet not handled (other destination)
}

void Process_sensorHandler(void* sdata) {
	sdata = 0;
}

void onAODVPacketReceived(aodv_dataPacket_t *dataPacket) {
	if (dataPacket->header.application == AODV_TEST_APPLICATION) {
		processTestAODVPacket(dataPacket);
		return;
	}
	testlog(LOGTAG, "Message arrived:");
	((char*) dataPacket->payload)[dataPacket->payload_size] = '\0';
	testlogm("[%s]", (char*) dataPacket->payload);
	logsep();
}
void onRREQReceived(aodv_RREQ_message_t *aRREQ) {
	static uint8_t blinkCount;
	if (aRREQ->destIPAddress == Configuration.id) {
		blinkCount = 5;
	}
	else {
		blinkCount = 2;
	}
	blinkNTimes(blinkCount);
}

void onPacketForward(aodv_dataPacket_t *forwardData) {
	// Check if data packet is an TEST_APPLICATION packet.
	if (forwardData->header.application != AODV_TEST_APPLICATION) {
		return;
	}
	// Convert to an AODV-Testpacket.
	aodv_test_packet_t *test_packet = (aodv_test_packet_t*) forwardData->payload;

	// Increase the hop-count.
	if (test_packet->isForwardDirection) {
		test_packet->forward_hopCount++;
	}
	else {
		test_packet->reverse_hopCount++;
	}
	blink(10);
}

void onAODVPacketSent(enum aodv_sendPacketStatus_e sendStatus) {
	switch (sendStatus) {
		case AODV_PACKET_SENT : {
			log(LOGTAG, "A Packet has been sent.");
			break;
		}
		case AODV_PACKET_NOT_SENT : {
			log(LOGTAG, "A Packet has NOT been sent.");
			break;
		}
		case AODV_DESTINATION_UNREACHABLE : {
			log(LOGTAG, "A packet has NOT been sent after the maximal number of retries (%i).", RREQ_RETRIES);
			break;
		}
		case AODV_BUFFER_FULL : {
			log(LOGTAG, "The packet could not be buffered, because the buffer is full.");
			break;
		}
	}
	logsep();
}


/**
 * Callback method, called by the timer which has been started
 * in the "init"-function.
 * When the LED goes off, that means that the MSP and its timer is
 * ready.
 */
void reportStarted(void) {
	LED_OFF;
	testlog(LOGTAG, "------------- The module is ready and tuned ;-) ----------------");
}

#define BUTTON_1 (BIT5)
#define BUTTON_2 (BIT6)
#define BUTTON_3 (BIT7)
#define BUTTONS (BUTTON_1 + BUTTON_2 + BUTTON_3)

uint8_t PORT1_FLAGS = 0x0;

void onButton1_pressed_handler(void) {
	log("Button-1", "RESET!");
	// LOL ;-) : WDTCTL = 0xDEAD;
	System_reset(WDS_SWRESET);
}

void onButton2_pressed_handler(void) {
//	sendConnectionTest();
}

void onButton3_pressed_handler(void) {

}

void interruptHandler(uint8_t *flags) {

	if (*flags & BUTTON_1) {
		onButton1_pressed_handler();
	}
	if (*flags & BUTTON_2) {
		onButton2_pressed_handler();
	}
	if (*flags & BUTTON_3) {
		onButton3_pressed_handler();
	}
	P1IE |= (BUTTON_1 + BUTTON_2 + BUTTON_3);

	*flags = 0x0;
}


interrupt(NOVECTOR) Port1_interrupt() {
	if ((P1IFG & BUTTONS) && !PORT1_FLAGS) {

		P1IE &= !(BUTTON_1 + BUTTON_2 + BUTTON_3);

		PORT1_FLAGS = P1IFG;

		P1IFG = 0;

		Timers_add(50, (fp_timer) interruptHandler, &PORT1_FLAGS);
	}
	P1IFG = 0;
}

void Process_init() {
	System_registerCallback( C_RADIO, ( fp_vp )Process_radioHandler );
	System_registerCallback( C_SENSOR, ( fp_vp )Process_sensorHandler );


	aodv_interface_initAODV();
	aodv_interface_addDataReceivedListener(onAODVPacketReceived);
	aodv_interface_addRREQReceivedListener(onRREQReceived);
	aodv_interface_addPacketForwardListener(onPacketForward);

/* Setup for Button-Interface... */
//	SET_ISR(PORT1_VECTOR, Port1_interrupt);
//	P1IFG = 0;
//	P1SEL = 0x0;
//	P1DIR = !(BIT5 + BIT6 + BIT7);
//	P1IES = !(BIT5 + BIT6 + BIT7);
//	P1IE = (BIT5 + BIT6 + BIT7);

	LED_ON;
	Timers_add(1, (fp_timer) reportStarted, NULL);
}


/******************* CONNECTION TEST START ****************************/
bool bConnectionTestStarted = false;
#define CONNECTION_TEST 					CON_TEST
#define CONNECTION_TEST_START 				CON_TEST_START
#define CONNECTION_TEST_STOP 				CON_TEST_STOP
#define CONNECTION_TEST_INTERVAL 			(2000)
#define CONNECTION_TEST_BLINK_INTERVAL 		(333)
#define _BROADCAST(_cmd_) 					("@255 "#_cmd_)
#define BROADCAST(_cmd_) 					_BROADCAST(_cmd_)
void sendConnectionTest(void) {
	if (!bConnectionTestStarted) {
		while (Timers_remove((fp_timer) sendConnectionTest, NULL));
		return;
	}
	_Cmd_AsciiSerialRequest(BROADCAST(CONNECTION_TEST));
	Timers_add(CONNECTION_TEST_INTERVAL, (fp_timer) sendConnectionTest, NULL);
}
COMMAND(CONNECTION_TEST, 0, cmdargs) {
	blink(CONNECTION_TEST_BLINK_INTERVAL);
	cmdargs = NULL;
}
COMMAND(CONNECTION_TEST_START, 0, cmdargs) {
	if (bConnectionTestStarted) {
		snprintf(cmdargs->response_args, cmdargs->response_bufsize, "Already started.");
		return;
	}
	bConnectionTestStarted = true;
	sendConnectionTest();
	snprintf(cmdargs->response_args, cmdargs->response_bufsize, "OK");
}
COMMAND(CONNECTION_TEST_STOP, 0, cmdargs) {
	bConnectionTestStarted = false;
	while (Timers_remove((fp_timer) sendConnectionTest, NULL));
	snprintf(cmdargs->response_args, cmdargs->response_bufsize, "OK");
}
/******************* CONNECTION TEST END ****************************/

/**
 * @brief	Ping request
 *
 * Send a ping request to another node
 * Example: ping 24
 */
COMMAND(ping, 0, cmdargs) {
	uint16_t node;

	node = String_parseUint16( cmdargs->args, NULL );
	if ( node > 0 ) {
		Net_sendPing((netaddr_t)node);
	}
}



uint8_t timesToBlink = 0;

COMMAND(blink, CMDFLAG_SERIAL, cmdargs) {
	uint16_t count;

	count = String_parseUint16(cmdargs->args, NULL);

	timesToBlink = count;
	blinkNTimes(timesToBlink);
}

COMMAND(AODV, CMDFLAG_SERIAL, cmdargs) {
	const char *data;
	uint16_t destination;
	uint8_t flags = AODV_FLAG_NONE, flags_length, i;
	uint16_t count = Cmd_SplitArgs(cmdargs->args);

	if (count < 2 || count > 3) {
		printf("[AODV]: Wrong number of parameters passed! (expected exactly 2)\r\n");
		printf("         Input format is:\r\n");
		printf("         AODV [<FLAGS>] <DEST> <DATA>\r\n");
		printf("              <FLAGS>: D|G|A\r\n");
		printf("                       D: Destination reply only.\r\n");
		printf("                       G: Use Gratuitous RREPs.\r\n");
		printf("                       A: Acknowledge RREPs.\r\n");
		return;
	}

	data = cmdargs->args;

	if (count == 3) {
		// Reading arg 1: Flags.
		flags_length = strlen(data);
		for (i = 0; i < flags_length; i++) {
			switch (data[i]) {
			case 'D' : flags |= AODV_FLAG_DESTONLY; break;
			case 'G' : flags |= AODV_FLAG_GRATUITOUS; break;
			case 'A' : flags |= AODV_FLAG_ACKRREPS; break;
			default: break;
			}
		}
		data = (char *) Cmd_GetNextArg(data);
	}

	// Reading arg 2: Destination (integer)
	destination = String_parseUint16(data, NULL);

	// Reading arg 3: Actual data (string)
	data = (char *) Cmd_GetNextArg(data);
	aodv_interface_sendData(destination, data, strlen(data), flags, 0, onAODVPacketSent);

}


/**
 * @brief Echo Arguments
 *
 * Testcase 1: long command name
 * Testcase 2: argument splitting
 */
COMMAND(echoargs, CMDFLAG_SERIAL, cmdargs) {
	uint16_t i;
	uint16_t count = Cmd_SplitArgs( cmdargs->args );
	char* arg = (char*) cmdargs->args;

	for ( i = 1; i <= count; i++ ) {
		printf("%u: %*s\n", i, strlen(arg), arg);
		arg = ( char* )Cmd_GetNextArg( arg );
	}
	snprintf(cmdargs->response_args, cmdargs->response_bufsize, "%u args found", count );
}

// Software header.
APP(
    APPH_STATIC_APP,	// Application Type
    __APP_NAME,			// Application Name (8 chars)
    1, 0				// Aplication Version (2 bytes)
);
