/*
 * aodv.genericList.h
 *
 *  Created on: 10.07.2009
 *      Author: Janos
 */

#ifndef AODV_GENERICLIST_H_
#define AODV_GENERICLIST_H_

#include <string.h>

#include <ScatterWeb.System.h>
#include <ScatterWeb.Net.h>

/**
 * @brief Generic structure for list elements.
 *
 * This structure defines a double linked list for
 * untyped pointer types. An element of this struct can
 * be a common list element, or define a list itself.
 *
 * Every list element can be the begin of a list or the
 * end of a list.
 */
typedef struct aodv_list_s aodv_list_t;
struct aodv_list_s {
	aodv_list_t *prev;
	aodv_list_t *next;
	void *data;
};

/**
 * @brief Creates an element in the list.
 *
 * This creates an element in the provided list 'aList' by
 * allocating memory for both the list element and the needed
 * storage space for the list elements data. This needed storage
 * is provided by the calle in the 'dataSize' argument.
 *
 * The function allocates memory for the list element and the
 * user data. The user data is linked to the list elements
 * data attribute.
 *
 * The new list element is added at the beginning of the list!
 * This provides the opportunity to insert an element in O(1)
 * operations.
 *
 * After the memory for both, the list element and the user data
 * is allocated, they are both cleared (using memset) by setting
 * to zero.
 *
 * Here is an example call:
 * aodv_list_t *my_list = NULL;
 * myType_t newItem = list_createEntry(&my_list, sizeof(myType_t));
 * -- "newItem" is now ready to use!
 *
 * @param **aList		A pointer to a lists first element. This element
 * 						is expected to be the beginning of a list, but
 * 						this expectation is NOT checked.
 * @param dataSize		The amount of memory (in bytes) which will be needed
 * 						for the created list element.
 * @return 				The pointer to the memory which has been allocated
 * 						for the user data, and which is now also held by
 * 						an element in the list.
 */
void* list_createEntry(aodv_list_t **aList, uint8_t dataSize);

aodv_list_t *list_createCustomEntry(aodv_list_t **aList, void *data_ptr);

/**
 * @brief Deletes an item from the list and frees the used memory.
 *
 * This searches for an list element, which holds the given
 * element pointer (by pointer comparison). If such an element
 * exists, it is removed from the list (by putting 'prev' and 'next'
 * together, and the memory of both the list element and the list
 * elements user data are freed (using free()).
 *
 * This function will only remove the first element which holds
 * the searched pointer.
 *
 * The function returnes TRUE, if an element was found and removed,
 * otherwise FALSE.
 *
 * (This function uses list_findEntry() to search for an Element).
 *
 * @param **aList		A pointer to a lists first element. This element
 * 						is expected to be the beginning of a list, but
 * 						this expectation is NOT checked.
 * @param *anEntry		A pointer, which is expected to be stored by a
 * 						list element. This list element is wanted to be
 * 						removed.
 * @param return		If an entry has been found and removed.
 */
bool list_deleteEntry(aodv_list_t **aList, void *anEntry);

/**
 * @brief Searches for an list element that holds the given pointer.
 *
 * This searches the list 'aList' for an element, which stores the
 * pointer 'anEntry' in it's 'data' attribute. If such an element is
 * found, it is returned.
 * Otherwise NULL is returned.
 *
 * As opposed to the functions 'delete' and 'create', the passed list
 * attribute 'aList' doesn't have to be the lists first element. The
 * element specified here can be the any element in the list, from
 * which the search is started (always using the NEXT reference).
 *
 * @param **aList		A pointer to a list element.
 * @param *anEntry		A pointer, which is expected to be stored by a
 * 						list element. This list element is wanted to be
 * 						removed.
 */
aodv_list_t* list_findEntry(aodv_list_t *aList, void *anEntry);

/**
 * @brief Deletes the complete list and frees all the memory that was allocated
 * for it. This will also set the pointer to the list, which is passed to thi
 * function, to NULL.
 *
 * This will delete every single entry in the list, but without the unnecessary
 * cost of joining the temporarily remaining list elements together. (This is
 * not necessary as these elements will be deleted anyway).
 *
 * For every list element, the allocated memory for the "data" attribute, and
 * the element memory itself is freed. Finally, the provided pointer to the
 * list is set to NULL, so that it doesn't point to unallocated memory.
 *
 * The 'aList' pointer is expected to be the head of the list.
 *
 * @param **aList		A pointer to a lists first element. This element
 * 						is expected to be the beginning of a list, but
 * 						this expectation is NOT checked.
 */
void list_clearList(aodv_list_t **aList);

/**
 * @brief Calculates the number of elements in the list.
 *
 * This calculates the number of elements in the list by traversing it. The
 * provided parameter 'aList' is expected, but doesn't have to be the head
 * of the list. If it is not, the number of successing elements are counted.
 *
 * @param *aList		A list or list element, from which the successors are
 * 						counted.
 *
 */
uint8_t list_getSize(aodv_list_t *aList);

#endif /* AODV_GENERICLIST_H_ */
