/*
 * aodv.interface.h
 *
 *  Created on: 20.05.2009
 *      Author: Janos
 */

#ifndef AODV_INTERFACE_H_
#define AODV_INTERFACE_H_

#include <string.h>

#include <ScatterWeb.System.h>
#include <ScatterWeb.Net.h>

#include "controller.netTypes.h"

#include "aodv.configuration.h"
#include "aodv.messageTypes.h"
#include "aodv.routing.table.h"
#include "aodv.genericList.h"


/**
 * Function definition for a handler function that will receive any incoming
 * data packages that are passed to the higher layers (>layer-3) of this
 * nodes application.
 *
 * The application MUST save the received data out of the package immediately.
 *
 * TODO: The Argument type should not be an AODV-data-packet, but a more general
 * object that contains only the payload, the payload-size and the originator
 * address.
 */
typedef void(*dataReceivedListener) (aodv_dataPacket_t *receivedData);

typedef void(*RREPReceivedListener) (aodv_RREP_message_t *receivedRREP);
typedef void(*RREQReceivedListener) (aodv_RREQ_message_t *receivedRREQ);

enum aodv_sendPacketStatus_e {
	AODV_PACKET_SENT,
	AODV_BUFFER_FULL,
	AODV_PACKET_NOT_SENT,
	AODV_DESTINATION_UNREACHABLE
};

/**
 * Function definition for a handler function that will be called by AODV to
 * inform applications ofer the success or failure of the sending of data
 * packets.
 */
typedef void(*aodv_applicationCallback) (enum aodv_sendPacketStatus_e sendStatus);

void aodv_interface_initAODV(void);

/*
 * Interface for the application to send data to a destination.
 *
 * The underlaying AODV protocol will take all necessary actions to
 * discover a route to the destination. The data will be stored until
 * a route is found and then transmitted.
 *
 * @param destination The destination "IP"-Address.
 * @param *data The payload data.
 * @param dataSize The size of the payload data in bytes.
 * @param flags Signals for AODV for the handling of possible required
 * 			route establishment. Possible flags are "G" and "D".
 * @param handler A callback handler method which is used by AODV to inform
 * 			the application over the success or failure of sending out the data packet.
 */
void aodv_interface_sendData(netaddr_t destination, const void *data, uint16_t dataSize, uint8_t flags, uint16_t application, aodv_applicationCallback handler);

/**
 * Sends an AODV-data frame to the destination 'destination'
 *
 * @param *aDataPacket The prepared data packet.
 * @param destination The "IP"-Address of the destination.
 *
 * @return TRUE or FALSE as a delegation of "ScatterWeb.Net.Net_send()".
 */
bool aodv_interface_sendDataFrame(aodv_dataPacket_t *aDataPacket, netaddr_t destination);

/**
 * Broadcasts a RREQ.
 *
 * @param *aRouteRequest A prepared RREQ,
 *
 * @return TRUE or FALSE as a delegation of "ScatterWeb.Net.Net_send()".
 */
bool aodv_interface_broadcastRREQPackage(aodv_RREQ_message_t *aRouteRequest);

/**
 * This will send the route reply RREP 'aRouteReply' towards the
 * next hop address 'nextHop'.
 *
 * @param *aRouteReply Pointer to a prepared RREP message.
 * @param nextHop "IP"-Address of the NEXT_HOP address, to which the
 * 			RREP will be sent.
 *
 * @return TRUE or FALSE as a delegation of "ScatterWeb.Net.Net_send()".
 */
bool aodv_interface_sendRREPPackage(aodv_RREP_message_t *aRouteReply, netaddr_t nextHopAddress);

/**
 * Registers a handler function that will receive every incoming data packet
 * from the layer-3 AODV protocol.
 *
 * @param aHandler A function which will be called by AODV for every received
 * 			data packet.
 */
bool aodv_interface_addDataReceivedListener(dataReceivedListener aListener);
bool aodv_interface_removeDataReceivedListener(dataReceivedListener aListener);
/**
 * Passes the supplied data packet through to the application using the
 * registered "onReceivedHandler", if one has been registered.
 *
 * @param The layer-3 packet that has been received.
 */
void aodv_interface_invokeDataReceivedListeners(aodv_dataPacket_t *receivedData);


bool aodv_interface_addRREPReceivedListener(RREPReceivedListener aListener);
bool aodv_interface_removeRREPReceivedListener(RREPReceivedListener aListener);
void aodv_interface_invokeRREPReceivedListeners(aodv_RREP_message_t *receivedRREP);
bool aodv_interface_addRREQReceivedListener(RREQReceivedListener aListener);
bool aodv_interface_removeRREQReceivedListener(RREQReceivedListener aListener);
void aodv_interface_invokeRREQReceivedListeners(aodv_RREQ_message_t *receivedRREQ);
bool aodv_interface_addPacketForwardListener(dataReceivedListener aListener);
bool aodv_interface_removePacketForwardListener(dataReceivedListener aListener);
void aodv_interface_invokePacketForwardListeners(aodv_dataPacket_t *receivedData);

bool sendPlainData(netaddr_t dest, void *data, size_t size);

#endif /* AODV_INTERFACE_H_ */
