/*
 * aodv.types.h
 *
 *  Created on: 12.05.2009
 *      Author: Janos
 */

#ifndef AODV_MESSAGE_TYPES_H_
#define AODV_MESSAGE_TYPES_H_


#include <ScatterWeb.System.h>
#include <ScatterWeb.Net.h>
#include <ScatterWeb.Configuration.h>

#include "aodv.genericList.h"

#define AODV_MESSAGE_TYPE_RREQ 			(1)
#define AODV_MESSAGE_TYPE_RREP 			(2)
#define AODV_MESSAGE_TYPE_RERR 			(3)
#define AODV_MESSAGE_TYPE_RREPACK 		(4)

#define AODV_FLAG_NONE 					(0x0)
#define AODV_FLAG_DESTONLY 				(0x01)
#define AODV_FLAG_GRATUITOUS 			(0x02)
#define AODV_FLAG_ACKRREPS 				(0x04)

/**
 * This enum specifies constant values as they are used in AODV control
 * messages. The definition is as follows:
 * 1 - AODV_RREQ
 * 2 - AODV_RREP
 * 3 - AODV_RERR
 * 4 - AODV_RREP-ACK
 */
enum aodv_controlMessages_e {
	TYPE_RREQ = 1,
	TYPE_RREP,
	TYPE_RERR,
	TYPE_RREPACK
};

/**
 * "Sort of" interface for "aodv_RREQ_message_t", "aodv_RREP_message_t" and
 * "aodv_RERR_message_t".
 *
 * As all of these structs share the "uint8_t type" attribut at the first
 * position in memory, a struct of these types can be casted to this
 * "interface"-struct to check which concrete type it belongs to.
 *
 * According to the value of the 'type' attribute the data can then be cast
 * to the correct struct type.
 */
typedef struct aodv_ControlMessage_s {
	uint8_t type;
} aodv_ControlMessage_t;

/**
 * "implements aodv_ControlMessage_s"
 */
typedef struct aodv_RREQ_message_s {
	/* Type = 1 */
	uint8_t type;

	/* Join flag, reserved for multicast */
	unsigned join :1;

	/* Repair flag, reserved for multicast */
	unsigned repair :1;

	/* Gratuitous RREP flag */
	unsigned gratuitous :1;

	/* Destination only flag */
	unsigned destOnly :1;

	/* Unknown sequence number flag */
	unsigned unknownSeqNum :1;

	/* Reserved.
	 * Sent as 0; ignored on reception. */
	unsigned reserved :3;

	/*
	 * TTL - Time To Live, used to control disemination of
	 * RREQ messages.
	 */
	uint8_t ttl;

	/* Hop Count.
	 * The number of hops from the Originator IP Adress
	 * to the node handling the request. */
	uint8_t hopCount;

	/* Request ID.
	 * A sequence number uniquely identifying the
	 * particular RREQ when taken in conjunction with the
	 * originating node's IP address. */
	uint16_t requestID;

	/* Destination IP Address. The IP address of the destination
	 * for which a route is desired. */
	netaddr_t destIPAddress;

	/* Originator IP Address. The IP address of the node which
	 * originated the Route Request. */
	netaddr_t origIPAddress;

	/* Destination Sequence Number.
	 * The latest sequence number received in the past by the
	 * originator for any route towards the destination. */
	uint16_t destSeqNum;

	/* Originator Sequene Number.
	 * The current sequence number to be used in the route entry
	 * pointing towards the originator of the route request. */
	uint16_t origSeqNum;

} aodv_RREQ_message_t;

/**
 * "implements aodv_ControlMessage_s"
 */
typedef struct aodv_RREP_message_s {
	/* Type = 2 */
	uint8_t type;

	/* Repair flag. Used for multicast. */
	unsigned repair :1;

	/* Acknowlegment required. */
	unsigned ackRequired :1;

	/* Reserved.
	 * Sent as 0; ignored on reception. */
	unsigned reserved: 9;

	/* Prefix Size.
	 * If nonzero, the 5-bit Prefix Size specifies that the indicated
	 * next hop max be used for any nodes with the same routing prefix
	 * (as defined by the prefix size) as the requested destination. */
	unsigned prefixSize :5;

	/* Hop Count.
	 * The number of hops from the Originator IP Address to the
	 * Destinantion IP Address. */
	uint8_t hopCount;

	/* Destinantion IP Address.
	 * The IP address of the destination for which a route is
	 * supplied. */
	netaddr_t destIPAddress;

	/* Originator IP Address.
	 * The IP addres of the node which originated the RREQ for
	 * which the route is supplied. */
	netaddr_t origIPAddress;

	/* Destinatio Sequence Number.
	 * The destination sequence number associated to the route. */
	uint16_t destSeqNum;

// XXX Lifetime of control messages are no longer considered.
//	/* Lifetime.
//	 * The time in milliseconds for which nodes receiving the
//	 * RREP consider the route to be valid. */
//	uint16_t lifetime;

} aodv_RREP_message_t;

/**
 * "implements aodv_ControlMessage_s"
 */
typedef struct aodv_RERR_message_s {
	/* Type = 3 */
	uint8_t type;

	/* No delete flag.
	 * Set when a node has performed a local repar of a link,
	 * and upstream nodes should not delete the route. */
	unsigned noDelete:1;

	/* Sent as 0; ignored on reception. */
	unsigned reserved : 15;

	/* DestCount.
	 * The number of unrechable destinantions included in the
	 * message; MUST be at least 1. */
	uint8_t destCount;

	/*
	 * Unreachable Destinantion IP Addresses.
	 * <Array>.
	 */
	netaddr_t *unreachableDestIPAddresses;

	/*
	 * Unreachable Destination Sequence Numbers.
	 * <Array>
	 */
    uint16_t *unreachableDestSeqNumbers;

} aodv_RERR_message_t;

/**
 * "implements aodv_ControlMessage_s"
 */
typedef struct aodv_RERPACK_message_s {
	/* Type = 4 */
	uint8_t type;
} aodv_RREPACK_message_t;

/**
 * Layer-3 header for AODV-data-packets.
 */
typedef struct aodv_header_s {
	netaddr_t destination;
	netaddr_t originator;
	uint8_t ttl;
	uint8_t flags;
	uint16_t application;
} aodv_header_t;

/**
 * Struct for wrapping data packages for AODV transmission.
 */
typedef struct aodv_dataPacket_s {
	aodv_header_t header;
	const void *payload;
	uint16_t payload_size;
} aodv_dataPacket_t;

#endif /* AODV_MESSAGE_TYPES_H_ */
