/*
 * aodv.routing.c
 *
 *  Created on: 13.05.2009
 *      Author: Janos
 */

#include "aodv.routing.h"


#define LOGTAG 			("routing")


aodv_dataPacket_t TEMP_DATA_PACKET;
aodv_RREQ_message_t TEMP_RREQ;
aodv_RREP_message_t TEMP_RREP;
aodv_RREP_message_t TEMP_HELLO_MESSAGE = {
		TYPE_RREP,
		0, 0, 0, 0, 0, 0, 0, 0
};

/***** PRIVATE METHODS *****/

/**
 * Extracts the AODV-Data-Frame out of the given NET-Frame-Payload.
 *
 * The AODV-data-frame is supplied with its AODV-header, the pointer that points
 * to the payload area within the frame and the correct size of the payload.
 *
 * @param aNetPacket (PTR) A layer-2 packet of the type "netTypes.h.userPackettypes_e.
 * 				AODV_DATA_PACKAGE" which was sent to this node.
 * @param anAodvDataPacket (PTR) A Pointer to a structure for saving the layer-3 data-frame
 * 				that is contained within the payload-data of the layer-2 NET-packet.
 */
void aodvExtractAODVDataPacket(struct netpacket_handler_args* aNetPacket, aodv_dataPacket_t *anAodvDataPacket);

/**
 * Handles the reception of a RREQ-packet.
 *
 * This includes updating and creating routes to the originator, previous
 * hop and destination of the RREQ, buffering the RREQ, and checking if
 * RREQ has affects this node itself.
 *
 * If the RREQ was sent to find this node, a RREP is created and sent. If
 * this node is not the searched node, the RREQ is being re-broadcasted.
 *
 * @param aNetPacket (PTR)A layer-2 packet of the type "netTypes.h.userPackettypes_e.
 * 				AODV_DATA_PACKAGE" which was received by this node.
 */
void processRREQ(struct netpacket_handler_args* aNetPacket);

/**
 * Handles the reception of a RREP-packet.
 *
 * @param aNetPacket (PTR)A layer-2 packet of the type "netTypes.h.userPackettypes_e.
 * 				AODV_DATA_PACKAGE" which was received by this node.
 */
void processRREP(struct netpacket_handler_args* aNetPacket);

/**
 * Checks the DATA_BUFFER for data packets, which can be sent now. Packets are
 * stored in the DATA_BUFFER because no valid ROUTING_TABLE entry could be
 * found for their destinations.
 * After the reception of a RREP message, the buffer should be checked though.
 *
 * If entries in the DATA_BUFFER are found, for which a valid routing table
 * entry exists, those packets are send out straight away. Those entries are
 * then marked as 'invalid' to avoid reprocessing.
 *
 * If entries in the DATA_BUFFER are found, which are timed out, they will be
 * marked as 'invalid'.
 */
inline void checkDataBufferForUpdates(void);

/**
 * Sends a data packet to an existent next hop entry from the ROUTING_TABLE.
 * The fact, that the next hop entry exists and is valid must be made
 * sure prior to this function call.
 */
void aodv_routeDataPacket(aodv_dataPacket_t *anAodvDataPacket, aodv_routingTableEntry_t *nextHopEntry);

/**
 * Sends a data packet from the data buffer to an existent next hop entry from the
 * ROUTING_TABLE. The fact, that the next hop entry exists and is valid must be
 * made sure prior to this function call.
 */
inline void aodv_routeBufferedDataPacket(aodv_packetBuffer_entry_t *packetBufferEntry,
		aodv_routingTableEntry_t *nextHopEntry);

/**
 * @brief Regenerates a RREQ as a n-th retry.
 *
 * This function is used in the Timer_add()-procedure, and the argument
 * will be a pointer to the buffered data packet in the PACKET_BUFFER.
 *
 * When this function is called by the timer, the buffered entry may already
 * have become inactive, e.g. because meanwhile a RREP has been received
 * and the buffered packet has already been sent.
 *
 * @param *arg 		The argument will be a pointer of the type "aodv_packetBuffer_entry_t*".
 * 					It points to the PACKET_BUFFER entry for which another RREQ will be
 * 					attempted.
 */
void regenerateRREQ(aodv_packetBuffer_entry_t* bufferEntry);

/**
 * @brief Prepares a RREQ for a buffered data packet and broadcasts it.
 *
 * This function creates a RREQ for a buffered PACKET_BUFFER entry and broadcasts
 * it. The possibly available routing table entry 'rtEntry' is used if a destination
 * sequence number is available for the destination of the data packet, but the route
 * is not active.
 *
 * After broadcasting the RREQ, a timed retransmission is initialized.
 *
 * @param *bufferEntry		An entry in the PACKET_BUFFER, for whiches destination a
 * 							route request is being established.
 * @param *rtEntry 			A route in the ROUTING_TABLE, whiches destination sequence
 * 							number will be used as the destination sequence number in the
 * 							RREQ, if the argument is NOT NULL.
 */
void prepareAndSendRREQ(aodv_packetBuffer_entry_t *bufferEntry, aodv_routingTableEntry_t *rtEntry);

/***************************/

/**
 * Any AODV control message (such as RREQ, RREP and RERR) has arrived
 * on layer-2 as either unicast to this node, unicast to other nodes or
 * as broadcast.
 */
void aodvProcessControlMessage(struct netpacket_handler_args* aNetPacket) {

	aodv_ControlMessage_t controlMessage;

	/* Construct a control-message struct from the payload of the layer-2 packet
	 * to determine of what control type the packet is. */
	memcpy(&(controlMessage), aNetPacket->payload, sizeof(aodv_ControlMessage_t));

	switch (controlMessage.type) {
	case TYPE_RREQ: {
		/* RREQ control message. Handle unicasts to >this< node and broadcasts. */
		if (aNetPacket->cpaResult == ADRCOMP_THIS || aNetPacket->cpaResult == ADRCOMP_BROADCAST) {
			log("layer-3", "RREQ received. Processing...");
			processRREQ(aNetPacket);
		}
		else {
			log("layer-3", "RREQ received, but for node %i. Ignoring it.", aNetPacket->netheader->to);
		}
		break;
	}
	case TYPE_RREP: {
		/* RREP control message. Handle unicasts to >this< node. */
		if (aNetPacket->cpaResult == ADRCOMP_THIS) {
			log("layer-3", "RREP received. Processing...");
			processRREP(aNetPacket);
			checkDataBufferForUpdates();
		}
		else if (aNetPacket->cpaResult == ADRCOMP_BROADCAST) {
			log("layer-3", "Broadcast RREP received.");
			aodv_processHELLO(aNetPacket, &TEMP_RREP);
			checkDataBufferForUpdates();
		}
		else {
			log("layer-3", "RREP received from %i, but for node %i. Ignoring it.", aNetPacket->netheader->from, aNetPacket->netheader->to);
		}
		break;
	}
	case TYPE_RERR: {

//		processRERR(aNetPacket);
		/*
		 * FIXME: Implement handling of RERRs.
		 * - RERRs can arrive as Unicast or Broadcasts.
		 *
		 * Receiving RERRs:
		 * - Procedure#1:
		 *   - Make LIST:
		 *   	- Add all destination from the RERR, for which a routing table
		 *   	  entry exists, which has the SENDER of the RERR as NEXTHOP.
		 *   - Procedure#3(LIST). (Prepare RERR).
		 *   - Update Routing table
		 *   	- For every destination in the RERR
		 *   		- If routing table exists for that destination
		 *   			- Copy the destination sequence number from the RERR into the entry
		 *   			- Invalidate the entry
		 *   			- Start deletion after DELET_PERIOD
		 *
		 * Creating RERRs:
		 * - Not more than RERR_RATELIMIT RERRs per second (will be skipped)
		 * - Procedure#2:
		 *   - Make LIST:
		 *   	- Add the unreachable destination.
		 *   	- Add destination, which uses the unrechable destination as a NEXTHOP
		 *   	  in the routing table.
		 *   - Procedure#3(LIST).
		 *   - Update Routing table
		 *   	- For every destination in the RERR
		 *   		- If routing table exists for that destination
		 *   			- Increase the destination sequence number.
		 *   			- Invalidate the entry
		 *   			- Start deletion after DELET_PERIOD
		 *
		 * - Procedure#3(LIST): Prepare RERR
		 * 	 - Call lists from Procedure#1 or Procedure#2 "LIST".
		 * 	 - Prepare RERR:
		 * 		- Add destinations:
		 * 			- All destinations from LIST, which have non-empty precursor-lists.
		 * 		- Create Receivers list "R":
		 * 			- Add all destinations from every PRECURSOR_LIST of all destinations
		 * 			  in the RERR "distinctly" into the list.
		 *
		 * - Procedure#4: Send out RERR
		 *   - Unicast or Broadcast RERR
		 *   	- Will be Unicast, if Precursors.size() == 1
		 *   	- Will be Broadcast, if Precursors.size() > 1
		 */
		log("layer-3", "RERR received. Handling not implemented yet...");
		break;
	}
	default : {
		log("layer-3", "Unknown packet received. Type-ID is '%i'.", controlMessage.type);
	}
	}
}



void broadcastHELLO(void) {
#ifndef DONT_SEND_HELLOS

	if (ROUTING_TABLE != NULL) {
		// Only send HELLOs, when a route exists in the routing table.
		TEMP_HELLO_MESSAGE.destIPAddress = Configuration.id;
		TEMP_HELLO_MESSAGE.destSeqNum = getOwnDestinationSequenceNumber();
		TEMP_HELLO_MESSAGE.hopCount = 0;

		log("HELLO", "Sending Hello message");

		aodv_interface_sendRREPPackage(&TEMP_HELLO_MESSAGE, NETADDR_BROADCAST);
	}
	// Restart the same thing.
	startSendHELLO_timer();
#endif
}

void startSendHELLO_timer(void) {
#ifndef DONT_SEND_HELLOS
	while (Timers_remove((fp_timer) broadcastHELLO, NULL));
	Timers_add(HELLO_INTERVAL, (fp_timer) broadcastHELLO, NULL);
#endif
}

void processRREQ(struct netpacket_handler_args* aNetPacket) {
	checkRREQResult_t processResult;

	/**
	 * Process the RREQ and get the result that determines the
	 * further processing.
	 */
	processResult = aodv_processRREQ(aNetPacket, &TEMP_RREQ);

	switch (processResult.result) {
	case RREQ_REBROADCAST: {
		log(LOGTAG, "RREQ must be re-broadcasted.");
		logm("Dest-Address=%i, Dest-Seq#=%i, RREQ-ID=%i.", TEMP_RREQ.destIPAddress, (TEMP_RREQ.unknownSeqNum ? -1 : (signed) TEMP_RREQ.destSeqNum), TEMP_RREQ.requestID);
		TEMP_RREQ.ttl--;
		logm("TTL: %i", TEMP_RREQ.ttl);
		if (TEMP_RREQ.ttl == 0) {
			log(LOGTAG, "RREQs TTL-value has reached 0 -> The message will not be retransmitted.");
			break;
		}
		aodv_interface_broadcastRREQPackage(&TEMP_RREQ);
		startSendHELLO_timer();

		aodv_interface_invokeRREQReceivedListeners(&TEMP_RREQ);

		break;
	}
	case RREQ_REACHED_DESTINATION: {
		log(LOGTAG, "RREQ has reached it's destination (this node).");
		logm("Preparing to reply with a RREP...");
		prepareRREPMessageFromRREQ(&TEMP_RREQ, &TEMP_RREP);
		log(LOGTAG, "RREP prepared:");
		logm("DEST:%i, Dest-Seq#: %i, ORIG: %i.", TEMP_RREP.destIPAddress, TEMP_RREP.destSeqNum, TEMP_RREP.origIPAddress);
		aodv_interface_sendRREPPackage(&TEMP_RREP, processResult.reverseRoute->nextHopAddress);

		aodv_interface_invokeRREQReceivedListeners(&TEMP_RREQ);

		break;
	}
	case RREQ_INTERMEDIATE_RREP:
	case RREQ_INTERMEDIATE_GRATUITOUS_RREP: {
		log(LOGTAG, "Found forward route towards the destination of the RREQ (%i).", TEMP_RREQ.destIPAddress);
		logm("Replying with intermediate RREQ.");
		prepareIntermediateRREPMessageFromRREQ(&TEMP_RREQ, &TEMP_RREP, processResult.forwardRoute);

		log(LOGTAG, "Intermediate RREP prepared:");
		logm("DEST:%i, Dest-Seq#: %i, ORIG: %i.", TEMP_RREP.destIPAddress, TEMP_RREP.destSeqNum, TEMP_RREP.origIPAddress);
		aodv_interface_sendRREPPackage(&TEMP_RREP, processResult.reverseRoute->nextHopAddress);

		if (processResult.result == RREQ_INTERMEDIATE_GRATUITOUS_RREP) {
			log(LOGTAG, "Preparing gratuitous RREP for destination '%i'.", processResult.forwardRoute->destAddress);
			prepareGratuitousRREPMessageFromRREQ(&TEMP_RREQ, &TEMP_RREP, processResult.reverseRoute);

			log(LOGTAG, "Gratuitous RREP prepared:");
			logm("DEST:%i, Dest-Seq#: %i, ORIG: %i.", TEMP_RREP.destIPAddress, TEMP_RREP.destSeqNum, TEMP_RREP.origIPAddress);
			aodv_interface_sendRREPPackage(&TEMP_RREP, processResult.forwardRoute->nextHopAddress);
		}
		break;
	}
	case RREQ_SELF_ORIGINATED: {
		log(LOGTAG, "The RREQ was originated by >this< node. It will be ignored.");
		break;
	}
	case RREQ_DUPLICATE: {
		log(LOGTAG, "The RREQ was already received earlier (buffered). It will be ignored.");
		break;
	}
	default:
		return;
	}
	startSendHELLO_timer();
}

void processRERR(struct netpacket_handler_args* aNetPacket) {
	aNetPacket = NULL;

//	aodv_RERR_message_t RERR;
//	uint8_t rerr_bytes = sizeof(aodv_RERR_message_t) - (2 * sizeof(void*));
//
//	memcpy(&RERR, aNetPacket->payload, rerr_bytes);
//	RERR.unreachableDestIPAddresses = aNetPacket->payload[rerr_bytes];
//	RERR.unreachableDestSeqNumbers = aNetPacket->payload[rerr_bytes + (sizeof(netaddr_t) * RERR.destCount)];
//
////	aodv_list_t *unreachableList = aodv_makeListFromRERR(aNetPacket, &RERR);

}

void processRREP(struct netpacket_handler_args* aNetPacket) {

	checkRREPResult_t processResult;

	processResult = aodv_processRREP(aNetPacket, &TEMP_RREP);

	switch (processResult.result) {
	case RREP_FORWARD: {
		log(LOGTAG, "Forwarding RREP over '%i'.",
				processResult.reverseRoute->destAddress);
		aodv_interface_sendRREPPackage(&TEMP_RREP, processResult.reverseRoute->destAddress);
		break;
	}
	case RREP_REACHED_DESTINATION: {
		log(LOGTAG, "RREP has reached it's destination (this node).");
		logm("Hop-Count: %i.", TEMP_RREP.hopCount);
		aodv_interface_invokeRREPReceivedListeners(&TEMP_RREP);
		break;
	}
	case RREP_NOROUTE: {
		log(LOGTAG, "ERROR: route for forwarding RREP found. Dropping RREP.");
		break;
	}
	default:
		;
	}

}

/**
 * @brief Attempts to find a route for a buffered data packet and route the data to
 * the found destination.
 *
 * This will perform a lookup for the destination of a buffered data packet. If a
 * valid route is found, this route will be used to transmit the data to it's
 * destination.
 * After transmitting the data, the buffered data will be set inactive and all
 * possible pending timer-events on this buffer entry will be removed.
 * The application callback handler will be called to inform the application that
 * the data has been sent out successfully.
 *
 * @param 	*dataPacketBufferEntry 	An entry in the PACKET_BUFFER, for which a transmission
 * 									is attempted.
 * @return 	A pointer to the route which has been used for the transmission, or the
 * pointer to the route which is not valid (and therefore
 */
inline aodv_routingTableEntry_t* attemptSendBufferedPacket(aodv_packetBuffer_entry_t *dataPacketBufferEntry) {
	aodv_routingTableEntry_t *rtEntry = aodv_routingTable_lookupRoute(dataPacketBufferEntry->packetHeader.destination);

	if (!rtEntry || !rtEntry->valid) {
		return rtEntry;
	}

	log(LOGTAG, "Found Buffered Data packet that can be sent now for destination %i.", dataPacketBufferEntry->packetHeader.destination);
	/**
	 * Removing all possible currently active Re-Request-timers on this
	 * buffered data packet.
	 */
	while (Timers_remove((fp_timer) regenerateRREQ, dataPacketBufferEntry));

	/**
	 * Send the data to it's destination.
	 */
	aodv_routeBufferedDataPacket(dataPacketBufferEntry, rtEntry);

	/**
	 * Invalidate the data buffer entry.
	 */
	if (dataPacketBufferEntry->callback_handler) {
		dataPacketBufferEntry->callback_handler(AODV_PACKET_SENT);
	}

	aodv_dropBufferedDataPacket(dataPacketBufferEntry);

	return rtEntry;
}

inline void checkDataBufferForUpdates(void) {

	aodv_list_t *packetBuffer = PACKET_BUFFER;

	/*
	 * Check the DATA_BUFFER for data packages that can now be sent
	 * since the reception of a RREP could have caused a route being
	 * established.
	 */
	while (packetBuffer != NULL) {
		attemptSendBufferedPacket(packetBuffer->data);
		packetBuffer = packetBuffer->next;
	}
}

void aodvExtractAODVDataPacket(struct netpacket_handler_args* aNetPacket, aodv_dataPacket_t *anAodvDataPacket) {
	memcpy(&(anAodvDataPacket->header), aNetPacket->payload, sizeof(aodv_header_t));
	anAodvDataPacket->payload = aNetPacket->payload + sizeof(aodv_header_t);
	anAodvDataPacket->payload_size = aNetPacket->payload_size - sizeof(aodv_header_t);
}

void aodvProcessDataPacket(struct netpacket_handler_args* aNetPacket) {

	aodv_dataPacket_t aodvPacket;

	enum netaddr_compareresult addressCompare;

	// Create/Update route to the previous hop
	aodv_routingTable_updateEntryFromIP(aNetPacket->netheader->from);

	aodvExtractAODVDataPacket(aNetPacket, &aodvPacket);

	// Create/Update route to the sender of the packet
	aodv_routingTable_updateEntryFromAODVPacket(&aodvPacket, aNetPacket);

	log("layer-3", "Extracted Packet: ORIG: '%i', DEST: '%i' (THIS: '%i').", aodvPacket.header.originator, aodvPacket.header.destination, Configuration.id);

	addressCompare = Net_compareAddress(&aodvPacket.header.destination, false);

	if (addressCompare == ADRCOMP_THIS || addressCompare == ADRCOMP_BROADCAST) {
		/*
		 * Package destined to >this< node.
		 * Pass it to the application.
		 */
		log(LOGTAG, "Packet reached this node.");

		aodv_interface_invokeDataReceivedListeners(&aodvPacket);

	} else if (addressCompare == ADRCOMP_OTHER) {
		/*
		 * Packet destined to another node.
		 * Routing actions must be taken.
		 */
		log(LOGTAG, "Packet must be forwarded to '%i'.", aodvPacket.header.destination);

		aodv_interface_invokePacketForwardListeners(&aodvPacket);

		aodvAttemptRouteDataPacket(&aodvPacket, 0x0, NULL);
	}

}

void regenerateRREQ(aodv_packetBuffer_entry_t* bufferEntry) {

	aodv_routingTableEntry_t *rtEntry;

	log(LOGTAG, "Attempting retransmission #%i for data packet for destination %i.", bufferEntry->numRetries, bufferEntry->packetHeader.destination);

	/**
	 * Check first if the buffered data can be sent.
	 */
	rtEntry = attemptSendBufferedPacket(bufferEntry);
	if (rtEntry && rtEntry->valid) {
		// The buffer entry has been sent (and is therefore now invalid).
		return;
	}
	// The buffer entry has not been sent and is still valid.

	if (bufferEntry->numRetries >= RREQ_RETRIES) {
		log(LOGTAG, "Maximum number of retries (%i) reached for data packet for destination %i.", RREQ_RETRIES, bufferEntry->packetHeader.destination);
		logm("The packet must be discarded.");
		if (bufferEntry->callback_handler) {
			bufferEntry->callback_handler(AODV_DESTINATION_UNREACHABLE);
		}
		aodv_dropBufferedDataPacket(bufferEntry);
		return;
	}

	log(LOGTAG, "Regenerating RREQ for data packet for destination %i: ", bufferEntry->packetHeader.destination);

	if (bufferEntry->nextTTL < TTL_THRESHOLD) {
		bufferEntry->nextTTL += TTL_INCREMENT;
	} else if (bufferEntry->nextTTL >= TTL_THRESHOLD) {
		bufferEntry->nextTTL = NET_DIAMETER;
		bufferEntry->numRetries++;
	}

	bufferEntry->nextBackoffTime = RING_TRAVERSAL_TIME(bufferEntry->nextTTL);

	prepareAndSendRREQ(bufferEntry, rtEntry);

	logsep();

}

void prepareAndSendRREQ(aodv_packetBuffer_entry_t *bufferEntry, aodv_routingTableEntry_t *rtEntry) {

	log(LOGTAG, "Preparing RREQ message for broadcasting.");
	prepareRREQMessage(&TEMP_RREQ, bufferEntry, rtEntry);

	log(LOGTAG, "RREQ prepared:");
	logm("Dest-Address=%i, Dest-Seq#=%i, RREQ-ID=%i.", TEMP_RREQ.destIPAddress, (TEMP_RREQ.unknownSeqNum ? -1 : (signed) TEMP_RREQ.destSeqNum), TEMP_RREQ.requestID);
	logm("TTL: %i, backoff: %i, retries: %i.", TEMP_RREQ.ttl, bufferEntry->nextBackoffTime, bufferEntry->numRetries);
	aodv_interface_broadcastRREQPackage(&TEMP_RREQ);
	startSendHELLO_timer();

	Timers_add(bufferEntry->nextBackoffTime, (fp_timer) regenerateRREQ, bufferEntry);

}

/**
 * @brief Generates and broadcasts a RREQ for an AODV-data-packet.
 *
 * The data packet itself will be buffered in the PACKET_BUFFER. The application callback handler, which
 * is used to inform the application over the progress of sending the data packet, is also stored in the
 * packet buffer.
 * The routing table entry, which is passed to this function, must be an invalid routing table entry for
 * the destination, or NULL. (If the routing table entry for the destination was valid, there would be no
 * need for sending a RREQ!). It is used to have a sequence number for the destination, if available, which
 * can be put into the RREQ.
 * The 'rreq_flags' are used to set the flags G and D in the RREQ message. Those flags are also buffered
 * with the data packet, so that they can be used again for a subsequent RREQ for the data packet.
 */
void generateRREQ(aodv_dataPacket_t *anAodvDataPacket, uint8_t rreq_flags, aodv_applicationCallback callback_handler,
		aodv_routingTableEntry_t *rtEntry) {

	aodv_packetBuffer_entry_t* bufferEntry;

	/* Buffer the data packet. */
	bufferEntry = aodv_bufferDataPacket(anAodvDataPacket, callback_handler, rreq_flags);
	if (bufferEntry == NULL) {
		/**
		 * Packet couldnt be buffered -> inform application.
		 */
		if (callback_handler) {
			callback_handler(AODV_BUFFER_FULL);
		}
		return;
	}

	prepareAndSendRREQ(bufferEntry, rtEntry);

}

bool aodvAttemptRouteDataPacket(aodv_dataPacket_t *anAodvDataPacket, uint8_t flags,
		aodv_applicationCallback callback_handler) {

	aodv_routingTableEntry_t *rtEntry;

	rtEntry = aodv_routingTable_lookupRoute(anAodvDataPacket->header.destination);

	log(LOGTAG, "Routing Datapacket to '%i'.", anAodvDataPacket->header.destination);

	if (!rtEntry || !rtEntry->valid) {
		/*
		 * No routing table entry was found:
		 * --> Packet must be buffered and RREQ must be established.
		 */
		log(LOGTAG, "No valid route found to '%i'.", anAodvDataPacket->header.destination);
		generateRREQ(anAodvDataPacket, flags, callback_handler, rtEntry);

		return false;
	}

	// Route found -> forward packet and done.
	log(LOGTAG, "Found route to '%i' over '%i'.", anAodvDataPacket->header.destination, rtEntry->nextHopAddress);

	aodv_routeDataPacket(anAodvDataPacket, rtEntry);


	/* Inform application that the packet has been sent. */
	if (callback_handler) {
		callback_handler(AODV_PACKET_SENT);
	}

	return true;
}

inline void aodv_routeBufferedDataPacket(aodv_packetBuffer_entry_t *packetBufferEntry,
		aodv_routingTableEntry_t *nextHopEntry) {

	log(LOGTAG, "Forwarding buffered data packet over '%i'.", nextHopEntry->nextHopAddress);

	TEMP_DATA_PACKET.payload = packetBufferEntry->bufferedData;
	TEMP_DATA_PACKET.payload_size = packetBufferEntry->data_size;
	memcpy(&TEMP_DATA_PACKET.header, &packetBufferEntry->packetHeader, sizeof(aodv_header_t));

	TEMP_DATA_PACKET.header.ttl--;

	aodv_interface_sendDataFrame(&TEMP_DATA_PACKET, nextHopEntry->nextHopAddress);
}

void aodv_routeDataPacket(aodv_dataPacket_t *anAodvDataPacket, aodv_routingTableEntry_t *nextHopEntry) {

	log(LOGTAG, "Forwarding packet over '%i'.", nextHopEntry->nextHopAddress);

	anAodvDataPacket->header.ttl--;

	aodv_interface_sendDataFrame(anAodvDataPacket, nextHopEntry->nextHopAddress);
}
