/*
 * aodv.routing.rrep.h
 *
 *  Created on: 03.06.2009
 *      Author: Janos
 */

#ifndef AODV_ROUTING_RREP_H_
#define AODV_ROUTING_RREP_H_

#include <string.h>

#include <ScatterWeb.System.h>
#include <ScatterWeb.Net.h>
#include <ScatterWeb.Configuration.h>
#include <ScatterWeb.CC1020.h>

#include "log.h"

#include "aodv.messageTypes.h"
#include "aodv.configuration.h"
#include "aodv.routing.table.h"
#include "aodv.interface.h"

/**
 * Enum values for naming some result values, which describe the
 * result of processing an AODV RREP message.
 */
typedef enum checkRREPResult {
	RREP_SELF_ORIGINATED, // value for signaling, that a RREP was originated by >this< node.
	RREP_NOROUTE, // value for signaling, that a forwarding route for a RREP couldn't be found.
	RREP_REACHED_DESTINATION, // value for signaling, that a RREP has reached it's destination at >this< node.
	RREP_FORWARD
// value for signaling, that a RREP can be forwarded.
} checkRREPResult_e;

/**
 * Type for storing a RREP-Check-Value together with a ROUTING_TABLE-entry.
 * Only if the result value is RREP_FORWARD, the 'nextRoute' value must
 * be set and may be read.
 * This attribute 'nextRoute' then MUST contain a valid route over which the
 * RREP can be forwarded towards it's origination.
 */
typedef struct checkRREPResult_s {
	checkRREPResult_e result;
	aodv_routingTableEntry_t *reverseRoute;
} checkRREPResult_t;

/**
 * @brief Processes a route reply (RREP) AODV message.
 *
 * This method receives a net-packet for which has been previously made sure
 * that it contains an AODV-RREP-control message in it's payload.
 *
 * This RREP message is processed and a result of the processing is returned.
 *
 * While processing, the ROUTING_TABLE will be updated towards the previous hop
 * and the originator of the RREP.
 *
 * According to the result value of this methode, a decision must be made how
 * to treat this RREP-message outside of this method call: If the RREP has reached
 * >this< node, it should not be forwarded. Instead, some actions can be resumed
 * which might have led to the process of discovering a route. If the RREP indeed
 * is not destined to this node, it should be forwarded. The result value in this
 * case contains a valid routing table enty, towards which the RREP must be
 * forwarded then.
 */
checkRREPResult_t aodv_processRREP(struct netpacket_handler_args* aNetPacket,
		aodv_RREP_message_t *preparedRREP);

/**
 * COMMENT: Documentation.
 */
void aodv_processHELLO(struct netpacket_handler_args* aNetPacket,
		aodv_RREP_message_t *preparedRREP);

/**
 * Prepares a passed RREP message as a response to a RREQ
 * for transmition towards the originator of the RREQ.
 *
 * The taken operations are:
 * 1. 	Increasing the own Sequence-Number by one, if the value
 * 		is then the Destination-Sequence-Number plus 1 in the RREQ.
 * 2. 	Writing the own Sequence-Number into the RREP.
 * 3. 	Setting the originators address from the RREQ as the
 * 		destination address in the RREP.
 * 4. 	Setting the own address into the RREP as its originator.
 * 5. 	Initializing the HOPCOUNT field with 0.
 * 6.	Setting the LIFETIME field to the value MY_ROUTE_TIMEOUT.
 */
void prepareRREPMessageFromRREQ(aodv_RREQ_message_t *aRREQ,
		aodv_RREP_message_t *preparedRREP);

/**
 * Prepares a passed RREP message as an intermediate RREP message for
 * transmission towards the originator of the RREQ, on which this is
 * the intermediate response.
 *
 * @param *aRREQ The RREQ message, for which this RREP is the reply.
 * @param *aRREP A pointer to a RREP, wich will be written to in this procedure.
 * @param *forwardRoute The route for the actual destination of the RREQ.
 */
void prepareIntermediateRREPMessageFromRREQ(aodv_RREQ_message_t *aRREQ,
		aodv_RREP_message_t *aRREP, aodv_routingTableEntry_t *forwardRoute);

/**
 * @brief Prepares a RREP message on the given pointer 'aRREP' for
 * transmission as a "gratuitous" reply to the destination of a RREQ,
 * when an intermediate RREP is sent to the originator.
 *
 * This conforms to RFC 3561, chapter 6.6.3.
 * A node sends an intermediate RREP back to the originator of a RREQ,
 * when ist has a valid route to the destination of a RREQ, and the
 * 'D' flag in the RREQ is not set.
 * When the 'G' flag is set in the RREQ, a forward RREP must be sent to
 * the destination of the RREQ, so the destination can learn of the
 * originator.
 * This RREP can be seen as the answer to a virtual RREQ, which was
 * originated by the destination of the RREQ.
 *
 * When preparing the RREP, the following steps are taken:
 * 1. The HOP_COUNT value is set to the value in the ROUTING_TABLE
 * entry for the originator of the RREQ.
 * 2. The destination IP address is set to the IP address of the
 * originator of the RREQ.
 * 3. The destination sequence number is set to the value for the
 * originator sequence number from the RREQ.
 * 4. The originator IP address is set to the IP address of the
 * destination of the RREQ (to whom this RREP will be sent).
 * 5. The lifetime field is set to the remaining lifetime of
 * the route towards the originator of the RREQ.
 *
 * @param *aRREQ A pointer to the RREQ that was received.
 * @param *aRREP A pointer to a RREP which can be modified. The memory
 * 		should be erased before using it.
 * @param *reverseRoute The pointer to the ROUTING_TABLE entry for
 * 		the originator of the RREQ.
 */
void prepareGratuitousRREPMessageFromRREQ(aodv_RREQ_message_t *aRREQ,
		aodv_RREP_message_t *aRREP, aodv_routingTableEntry_t *reverseRoute);

#endif /* AODV_ROUTING_RREP_H_ */
