/*
 * aodv.routingTable.c
 *
 *  Created on: 21.05.2009
 *      Author: Janos
 */

#include "aodv.routing.table.h"

#define LOGTAG ("routingtable")

/**************** PRIVATE METHODS ****************/
void deactivateRoute(aodv_routingTableEntry_t *anEntry);
void deleteRoute(aodv_routingTableEntry_t *anEntry);
/*************************************************/


aodv_list_t *ROUTING_TABLE = NULL;
uint8_t SIZEOF_AODVROUINGTABLEENTRY_T = sizeof(aodv_routingTableEntry_t);
uint8_t SIZEOF_AODVPRECURSOR_T = sizeof(aodv_precursor_t);


void deleteRoute(aodv_routingTableEntry_t *anEntry) {
	if (anEntry == NULL) {
		return;
	}
	log(LOGTAG, "Deleting route to destination %i.", anEntry->destAddress);
	// remove all pending timers for deleting this route.
	while (Timers_remove((fp_timer) deleteRoute, anEntry));
	// remove all pending timers for deactivating this route.
	while (Timers_remove((fp_timer) deactivateRoute, anEntry));

	list_clearList(&anEntry->precursor_list);

	list_deleteEntry(&ROUTING_TABLE, anEntry);
}

void deactivateRoute(aodv_routingTableEntry_t *anEntry) {
	if (anEntry == NULL) {
		return;
	}
	log(LOGTAG, "Deactivating route to destination %i.", anEntry->destAddress);
	anEntry->valid = false;
	// remove all pending timers for deleting this route.
	while (Timers_remove((fp_timer) deleteRoute, anEntry));

	Timers_add(DELETE_PERIOD, (fp_timer) deleteRoute, anEntry);
}

inline void startRouteDeactivationTimer(aodv_routingTableEntry_t *anEntry) {
	// remove all pending timers for deactivating this route.
	while (Timers_remove((fp_timer) deactivateRoute, anEntry));

	// Start the timer for deleting this route.
	Timers_add(MY_ROUTE_TIMEOUT, (fp_timer) deactivateRoute, anEntry);
}

void aodv_routingTable_reactivateRoute(aodv_routingTableEntry_t *anEntry) {

	if (anEntry == NULL) {
		return;
	}
	log(LOGTAG, "Reactivating route for destination %i.", anEntry->destAddress);

	anEntry->valid = true;

	// remove all pending timers for deleting this route.
	while (Timers_remove((fp_timer) deleteRoute, anEntry));

	startRouteDeactivationTimer(anEntry);
}



aodv_routingTableEntry_t* aodv_routingTable_lookupRoute(netaddr_t aDestination) {
	aodv_list_t *list = ROUTING_TABLE;
	aodv_routingTableEntry_t *entry;

	while (list != NULL) {
		entry = list->data;
		if (entry->destAddress == aDestination) {
			return entry;
		}
		list = list->next;
	}
	return NULL;
}

aodv_routingTableEntry_t* aodv_routingTable_createEntry(netaddr_t aDestination) {

	aodv_routingTableEntry_t *newEntry;

	newEntry = list_createEntry(&ROUTING_TABLE, SIZEOF_AODVROUINGTABLEENTRY_T);

	if (newEntry == NULL) {
		log(LOGTAG, "FATAL: No memory for routing table entry for destination %i could be allocated!", aDestination);
		return NULL;
	}

	log(LOGTAG, "Routing table entry created for destination %i.", aDestination);

	newEntry->destAddress = aDestination;
	newEntry->valid = true;
	newEntry->validDestSeqNumber = false;
	newEntry->precursor_list = NULL;

	return newEntry;
}

aodv_routingTableEntry_t* aodv_routingTable_updateEntryFromAODVPacket(aodv_dataPacket_t *aodvPacket, struct netpacket_handler_args* aNetPacket) {
	aodv_routingTableEntry_t* entry = aodv_routingTable_lookupRoute(aodvPacket->header.originator);
	log(LOGTAG, "Updating entry for destination '%i'...", aodvPacket->header.originator);

	if (entry == NULL) {
		logm("-> Entry doesn't exist. Create one...");
		entry = aodv_routingTable_createEntry(aodvPacket->header.originator);
		if (entry == NULL) {
			logm("---------> FATAL: Routing table entry couldn't be created!");
			return NULL;
		}
		startRouteDeactivationTimer(entry);
	}
	else {
		if (entry->nextHopAddress == aNetPacket->netheader->from) {
			/* reactivate route when the packet is received from the
			 * already known nextHop.
			 * Otherwise, no changes can be made to the existing route,
			 * because there is no information available.
			 */
			aodv_routingTable_reactivateRoute(entry);
		}
		return entry;
	}

	entry->hopCount = 255;
	entry->validDestSeqNumber = false;
	entry->nextHopAddress = aNetPacket->netheader->from;
	logm("-> Entry for destination '%i' updated.", aodvPacket->header.originator);

	return entry;
}

aodv_routingTableEntry_t* aodv_routingTable_updateEntryFromIP(netaddr_t aDestination) {
	aodv_routingTableEntry_t* entry = aodv_routingTable_lookupRoute(aDestination);
	log(LOGTAG, "Updating entry for destination '%i'...", aDestination);

	if (entry == NULL) {
		logm("-> Entry doesn't exist. Create one...");
		entry = aodv_routingTable_createEntry(aDestination);
		if (entry == NULL) {
			logm("---------> FATAL: Routing table entry couldn't be created!");
			return NULL;
		}
		startRouteDeactivationTimer(entry);
	}
	else {
		// reactivate route
		aodv_routingTable_reactivateRoute(entry);
	}

	entry->hopCount = 1;
	entry->validDestSeqNumber = false;
	entry->nextHopAddress = aDestination;
	logm("-> Entry for destination '%i' updated.", aDestination);

	return entry;
}

aodv_routingTableEntry_t* aodv_routingTable_updateForwardRouteFromRREP(
		aodv_RREP_message_t *aRREP, netaddr_t thePreviousHop) {

	int16_t seqNumCompareResult;
	aodv_routingTableEntry_t* entry;

	log(LOGTAG, "Checking forward route for destination '%i'...",
			aRREP->destIPAddress);

	entry = aodv_routingTable_lookupRoute(aRREP->destIPAddress);
	aRREP->hopCount++;

	if (entry == NULL) {
		logm("-> Entry doesn't exist. Create one...");
		entry = aodv_routingTable_createEntry(aRREP->destIPAddress);
	}
	else {
		aodv_routingTable_reactivateRoute(entry);
		seqNumCompareResult = ((signed) aRREP->destSeqNum) - ((signed) entry->destSeqNumber);
		if (!(!entry->validDestSeqNumber
				|| (seqNumCompareResult > 0)
				|| (seqNumCompareResult == 0 && !entry->valid)
				|| (seqNumCompareResult == 0 && aRREP->hopCount < entry->hopCount))) {
			// Routing table entry is up to date :) !
			logm("-> Route is up to date!");
			return entry;
		}
	}

	entry->validDestSeqNumber = true;
	entry->nextHopAddress = thePreviousHop;
	entry->hopCount = aRREP->hopCount;
	entry->destSeqNumber = aRREP->destSeqNum;

	logm("-> Route updated: New DestSeq#: %i.", aRREP->destSeqNum);
	logm("   Next-Hop is: '%i'.", entry->nextHopAddress);

	return entry;
}

/*
 * 3. 	Create or update reverse route to the ORIG_IP.
 * 		3.1. 	Originator Sequence Number from the RREQ is compared to the corresponding
 * 				destination sequence number in the route table entry and copied if greater
 * 				than the existing value.
 * 		3.2.	The valid sequence number field is set to TRUE.
 * 		3.3.	The next hop in the routing table becomes the node from which the RREQ was
 * 				received.
 * 		3.4.	The hop count is copied from the hopcount in the RREQ message.
 * 		3.5.	Lifetime is set to be the maximum of (ExistingLifetime, MinimalLifetime),
 * 				where 		MinimalLifetime	 =	(current time + 2*NET_TRAVERSAL_TIME -
 * 												 2*HopCount*NODE_TRAVERSAL_TIME).
 *
 * 4.	Update RREQ:
 * 		4.1.	TTL field in the IP header is decreased by one.
 * 		4.2. 	Increment the hop count value by one.
 * 		4.3.	Update the destination sequence number to the maximum of the current value
 * 				and the destination sequence number maintained by this node.
 */
aodv_routingTableEntry_t* aodv_routingTable_updateReverseRouteFromRREQ(
		aodv_RREQ_message_t *aRREQ, netaddr_t thePreviousHop) {

	int16_t seqNumCompareResult;
	aodv_routingTableEntry_t* entry;

	log(LOGTAG, "Updating reverse route to destination '%i'.",
			aRREQ->origIPAddress);

	entry = aodv_routingTable_lookupRoute(aRREQ->origIPAddress);

	if (entry == NULL) {
		logm("Entry doesn't exist. Create one...");
		entry = aodv_routingTable_createEntry(aRREQ->origIPAddress);
	}
	else {
		aodv_routingTable_reactivateRoute(entry);
	}

	/* Increment the hop count value */
	aRREQ->hopCount++;

	log(LOGTAG, "Updating sequence numbers:");
	logm("RREQ.origSeq#: %i", aRREQ->origSeqNum);

	if (entry->validDestSeqNumber) {
		/* DestSeqNum in RT is valid.
		 * --> COMPARE them using signed arithmetic. */
		seqNumCompareResult = ((signed) aRREQ->origSeqNum) - ((signed) entry->destSeqNumber);
		if (seqNumCompareResult > 0) {
			/* DestSeqNum in RREQ is greater:
			 * --> UPDATE routingTable entry. */
			entry->destSeqNumber = aRREQ->origSeqNum;
			log(LOGTAG, "Sequence number updated.");
		}
	}
	else {
		/* DestSeqNum in RT is invalid.
		 * --> UPDATE routingTable entry. */
		entry->destSeqNumber = aRREQ->origSeqNum;
		log(LOGTAG, "Sequence number for '%i' updated to '%i'.", entry->destAddress, entry->destSeqNumber);
	}
	entry->nextHopAddress = thePreviousHop;
	entry->hopCount = aRREQ->hopCount;
	entry->validDestSeqNumber = true;

	return entry;
}

void aodv_routingTable_addPrecursor(aodv_routingTableEntry_t *anEntry, netaddr_t aPrecursor) {

	aodv_precursor_t *tmp;
	aodv_list_t *tmpList;

	// Check if the precursor doesn't already exist in the precursor list.
	tmpList = anEntry->precursor_list;
	while (tmpList != NULL) {
		if (((aodv_precursor_t*) tmpList->data)->precursor == aPrecursor) {
			// Exit, if the precursor already exists.
			return;
		}
	}

	log(LOGTAG, "Adding node '%i' into the precursor list of destination %i.", aPrecursor, anEntry->destAddress);

	// Create an entry for the precursor.
	tmp = list_createEntry(&anEntry->precursor_list, SIZEOF_AODVPRECURSOR_T);

	if (tmp == NULL) {
		log(LOGTAG, "FATAL: Failed to add precursor %i to routing table entry for destination %i.", aPrecursor, anEntry->destAddress);
		return;
	}

	tmp->precursor = aPrecursor;
}
