/*
 * aodv.routing.table.h
 *
 *  Created on: 21.05.2009
 *      Author: Janos
 */

#ifndef AODV_ROUTING_TABLE_H_
#define AODV_ROUTING_TABLE_H_

#include <string.h>

#include <ScatterWeb.System.h>
#include <ScatterWeb.Net.h>
#include <ScatterWeb.Timers.h>

#include "log.h"

#include "aodv.messageTypes.h"
#include "aodv.configuration.h"
#include "aodv.genericList.h"

extern aodv_list_t *ROUTING_TABLE;

/**
 * Precursor list with nodes, which need to be informed about route
 * breakdowns.
 */
typedef struct aodv_precursor_s aodv_precursor_t;
struct aodv_precursor_s {
	netaddr_t precursor;
};

/**
 * Routing table entry for a destination node.
 */
typedef struct aodv_routingTableEntry_s aodv_routingTableEntry_t;
struct aodv_routingTableEntry_s {

	/* Destination IP Address */
	netaddr_t destAddress;

	/* Next Hop */
	netaddr_t nextHopAddress;

	/* Destination Sequence Number */
	uint16_t destSeqNumber;

	/* Valid Destination Sequence Number flag */
	unsigned validDestSeqNumber :1;

	unsigned valid :1;

	/* Other state and routing flags (e.g., valid, invalid, repairable, being repaired) */
	unsigned otherFlags :6;

	/* Hop Count (number of Hops needed to reach destination) */
	uint8_t hopCount;

	/* List of Precursors */
	aodv_list_t *precursor_list;

	/* Lifetime (expiration or deletion time of the route) */
	uint16_t lifetime;

};

/**
 * @brief Searches an entry in the ROUTING_TABLE, whos destination value equals
 * 'aDestination'.
 *
 * This searches the linked list of routing table entries subsequently. If an entry
 * is found in the list, that has the searched value as its destination value, the
 * pointer to that routing table entry is returned. Otherwise NULL is returned.
 *
 * @param aDestination 	A value for a nodes address, to whom a route is searched.
 * @return 				The routing table entry for the node that is searched, or
 * 						NULL, if such an entry does not exist.
 */
aodv_routingTableEntry_t* aodv_routingTable_lookupRoute(netaddr_t aDestination);

/**
 * Creates or updates a route to a nearby node.
 *
 * This nearby-node was discovered through a level-2 packet which was received.
 * The node therefore must be in direct connection to this node, and will obtain
 * a routing table entry with HOPCOUNT=1.
 *
 * Because this node was discovered through a level-2 packet, no destination sequence
 * number is available. Thus the route is marked as invalid.
 */
aodv_routingTableEntry_t* aodv_routingTable_updateEntryFromIP(netaddr_t aDestination);

/**
 * TODO: Documentation
 */
aodv_routingTableEntry_t* aodv_routingTable_updateEntryFromAODVPacket(aodv_dataPacket_t *aodvPacket, struct netpacket_handler_args* aNetPacket);

/**
 * TODO: Documentation.
 */
aodv_routingTableEntry_t* aodv_routingTable_updateForwardRouteFromRREP(aodv_RREP_message_t *aRREP, netaddr_t thePreviousHop);

/**
 * TODO: Documentation.
 */
aodv_routingTableEntry_t* aodv_routingTable_updateReverseRouteFromRREQ(aodv_RREQ_message_t *aRREQ, netaddr_t thePreviousHop);

/**
 * @brief Adds a precursor address into a routing table entries precursor list.
 *
 * The given address is added into the precursor list of the routing table
 * entry 'anEntry'.
 *
 * @param anEntry 		A routing table entry, into which the node address is added.
 * @param aPrecursor 	A node address to be inserted as a precursor into the routing table entry.
 */
void aodv_routingTable_addPrecursor(aodv_routingTableEntry_t *anEntry, netaddr_t aPrecursor);

/**
 * @brief Reactivates a route in the ROUTING_TABLE.
 *
 * TODO: Documentation...
 */
void aodv_routingTable_reactivateRoute(aodv_routingTableEntry_t *anEntry);

#endif /* AODV_ROUTING_TABLE_H_ */
