/*
Copyright 2007, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2007
*/

/**
 * @file			ScatterWeb.ADDA.c
 * @ingroup	libADDA
 * @brief			ADDA library
 * 
 * @author		Bastian Blywis
 * @date			Mi Okt 31 11:59:35 2007
 * @since		1.2
 * @version	$Revision$
 * 
 */

#include "ScatterWeb.ADDA.internals.h"
#include "ScatterWeb.ADDA.h"
#include "ScatterWeb.ADC.internals.h"

/******************************************************************************
 * @name	Initialization and configuration
 * @{
 */

/*
 * Inits the AD to contionously calc the 4 values (extern and battery voltage,
 * receivepower and microphone) and generate an interrupt at the end of this sequence,
 * the interrupt routine then copies out all values into ...Value
 */
void _ADDA_initADC(void) {

	/*
	 * SHT0_15: Sample & Hold Time 0 (ADC12MEM0 to ADC12MEM7) = 0xF
	 * SHT1_15: Sample & Hold Time 1 (ADC12MEM8 to ADC12MEM15) = 0xF
	 * MSC: multiple sample and conversion
	 */
	ADC12CTL0 = SHT0_15 + SHT1_15 + MSC;

	/*
	 * SHP: Sample-and-hold pulse-mode select = SAMPCON signal is sourced from the sampling timer
	 * CONSEQ_3: Conversion sequence mode select = Repeat-sequence-of-channels
	 * CSTARTADD_0: Start conversion at ADC12MEM0
	 */
	ADC12CTL1 = SHP + CONSEQ_3 + CSTARTADD_0;

	/*
	 * INCH_0: Input channel select = A0
	 * SREF_0: Select reference = VR+ = AVCC and VR- = AVSS
	 */
	ADC12MCTL0 = (INCH_0 + SREF_0);
	ADC12MCTL1 = (INCH_0 + SREF_0);
	ADC12MCTL2 = (INCH_0 + SREF_0);

	/*
	 * INCH_1: Input channel select = A1
	 * SREF_0: Select reference = VR+ = AVCC and VR- = AVSS
	 */
	ADC12MCTL3 = (INCH_1 + SREF_0);
	ADC12MCTL4 = (INCH_1 + SREF_0);
	ADC12MCTL5 = (INCH_1 + SREF_0);

	/*
	 * INCH_2: Input channel select = A2
	 * SREF_0: Select reference = VR+ = AVCC and VR- = AVSS
	 */
	ADC12MCTL6 = (INCH_2 + SREF_0);
	ADC12MCTL7 = (INCH_2 + SREF_0);
	ADC12MCTL8 = (INCH_2 + SREF_0);
	ADC12MCTL9 = (INCH_2 + SREF_0);

	/*
	 * INCH_3: Input channel select = A3
	 * SREF_0: Select reference = VR+ = AVCC and VR- = AVSS
	 */
	ADC12MCTL10 = (INCH_3 + SREF_0);

	/*
	 * INCH_4: Input channel select = A4
	 * SREF_0: Select reference = VR+ = AVCC and VR- = AVSS
	 */
	ADC12MCTL11 = (INCH_4 + SREF_0);

	/*
	 * INCH_5: Input channel select = A5
	 * SREF_0: Select reference = VR+ = AVCC and VR- = AVSS
	 */
	ADC12MCTL12 = (INCH_5 + SREF_0);

	/*
	 * INCH_10: Input channel select = Temperature sensor
	 * SREF_0: Select reference = VR+ = AVCC and VR- = AVSS
	 * EOS: End of sequence. Indicates the last conversion in a sequence.
	 */
	ADC12MCTL13 = (INCH_10 + SREF_1);

	/*
	 * INCH_15: Input channel select = (AVCC - AVSS) / 2
	 * SREF_1: VR+ = VREF+ and VR- = AVSS, VREF+ = 2.5V
	 */
	ADC12MCTL14 = (INCH_15 + SREF_1) + EOS;
}

bool ADDA_enableADC() {
	uint8_t on = false;
	if(ADC12CTL0 & ADC12ON)
		on = true;
	ADC12CTL0 |= ADC12ON;	// Switch the ADC on
	ADC12CTL0 |= ENC;		// enable conversion
	ADC12CTL0 |= ADC12SC;	// Start conversion

	LOG_HIGH("ADC: enabled");
	return on;
}

bool ADDA_disableADC() {
	uint8_t on = false;
	if(ADC12CTL0 & ADC12ON)
		on = true;

	ADC12CTL0 &= ~ENC; 		// Stop conversion
	while(ADC12CTL1 & ADC12BUSY) // usually only 3 cycles
		;
	ADC12CTL0 &= ~ADC12ON;	// turn off ADC12

	LOG_HIGH("ADC: disabled");
	return on;
}

/*
 * Read the temperature via the internal diode.
 * U = URef/4095 = 0.0006105006105006105V, URef=2.5V
 * Temperature = (U-0.986)/0.00355
 */
float ADDA_readTemperature() {
	uint8_t on = ADDA_disableADC();
	ADC12CTL0 |= REFON + REF2_5V;	// Switch reference on
	ADDA_enableADC();
	System_wait(ADDA_WAIT_TICKS);	// Delay for reference start-up

	float temp = 0.0006105006105006105 * VTMP;
	temp = (temp-0.986)/0.00355;

	ADDA_disableADC();
	ADC12CTL0 &= ~REFON + ~REF2_5V;	// Switch reference off

	if(on)
		ADDA_enableADC();

	return temp;
}

void ADDA_updateVeRef() {
	uint8_t on = ADDA_disableADC();
	ADC12CTL0 |= REFON + REF2_5V;	// Switch reference on
	ADDA_enableADC();
	System_wait(ADDA_WAIT_TICKS);	// Delay for reference start-up

	float volt = 0.0006105006105006105 * VCC * 2;
	ADDA_VeRef = volt;// * 0.9375; // correcture factor

	ADDA_disableADC();
	ADC12CTL0 &= ~REFON + ~REF2_5V;	// Switch reference off

	if(on)
		ADDA_enableADC();

	Timers_add(ADDA_VREF_S * 1024, (fp_timer) ADDA_updateVeRef, (void*) 0xFFFF);
}

void ADDA_initADC() {
	_ADDA_initADC();
	ADDA_VeRef = 3.0;
	ADDA_updateVeRef();
	ADDA_readTemperature();
	printf("ADC initialized\n");
}

float ADDA_bits2Volt(uint16_t bits) {
	float volt = ADDA_VeRef/4096 * bits;
	return volt;
}

float ADDA_readADC(uint8_t adc) {
	uint8_t on = ADDA_enableADC();
	switch(adc) {
		case 4:
			return ADDA_bits2Volt(AIN4);
		case 5:
			return ADDA_bits2Volt(AIN5);
		default:
			return -1;
	}
	if(!on)
		ADDA_disableADC();
}

void ADDA_printFloat(float f) {
	uint16_t pre = (uint16_t) f;
	uint16_t post = (f-pre) * 1000;

	if(f < 0)
		printf("-");
	printf("%u.", pre);
	uint16_t t = post;
	if(post) {
		while(t <= 100) {
			t*=10;
			post/=10;
			printf("0");
		}
	}
	printf("%u", post);
}

float ADDA_parseFloat(char* a, uint8_t* b) {
	float f = 0.0;
	while(*a != 0x00) {
		if(*a == '.') {
			a++;
			break;
		}
		f*=10;
		f+=*a-0x30;
		a++;
	}
	ADDA_printFloat(f);
	printf("\n");
	uint16_t t = 1;
	while(*a != 0x00) {
		t*=10;
		f+=(float)(*a-0x30)/t;
		a++;
	}
	ADDA_printFloat(f);
	printf("\n");
	if(b)
		b = a;
	return f;
}

/**
 * @brief Read current temperature:  <code>temp_int</code>
 * @ingroup commands
 * @return temperature in [C]
 * @see ScatterWeb.ADDA.h
 *
 * <p>Measures current temperature via the MSP430 internal temperature diode.</p>
 * <p>Use the more precise SHT11 if present on the MSB-430.</p>
 */
COMMAND(temp_int, 0, cmdargs) {
	printf("Temperature = ");
	ADDA_printFloat(ADDA_readTemperature());
	printf("C\r\n");
}

/**
 * @brief Read the battery voltage:  <code>battery</code>
 * @ingroup commands
 * @return battery voltage in [V]
 * @see ScatterWeb.ADDA.h
 *
 * Return the periodically measured battery voltage.
 */
COMMAND(battery, 0, cmdargs) {
	printf("Battery = ");
	ADDA_printFloat(ADDA_VeRef);
	printf("V\r\n");
}

/**
 * @brief Read voltage:  <code>readADC [ADC]</code>
 * @ingroup commands
 * @param ADC out of {4,5}
 * @return measured voltage in [V]
 * @see ADDA_readADC
 * @see ScatterWeb.ADDA.h
 *
 * Read ADC memory register.
 */
COMMAND(readADC, 0, cmdargs) {
	uint8_t adc = String_parseUint16(cmdargs->args, NULL);
	float volt = ADDA_readADC(adc);
	if(volt < 0)
		printf("ERROR\r\n");
	else {
		printf("ADC(%u) = ", adc);
		ADDA_printFloat(volt);
		printf("V\r\n");
	}
}

/** @} */

