#include "fat.internals.h"

/**
 * Reads a number of bytes from a file given by a handle.
 * This function cannot read across cluster borders!
 * 
 * @param	hFile		pointer to handle to file to read from
 * @param	pBuffer		buffer to read into
 * @param	count		number of bytes to read
 * @return	uint16_t	number of bytes read
 * 
 * @see Fat_FindFile_Simple
 */
uint16_t Fat_ReadFile(
	const fat_state_t*	pFS, 
	file_handle_t*		hFile,
	void*				pBuffer,
	uint16_t				count
) {
	uint16_t ret;
	uint32_t address;
	
	address = _Fat_GetFilePos( pFS, hFile );
	ret = Sd_read( pBuffer, address, count );
		
	File_Seek( pFS, hFile, ret, NULL );
	
	return ret;
}

/**
 * @name Handle handling
 */
void File_Close( file_handle_t* hFile ) {
	memset(hFile, 0, sizeof(file_handle_t));	// just like new
}

void File_Reset( file_handle_t* hFile ) {
	hFile->CurrentCA = hFile->StartCA;
	hFile->Position = 0;
	hFile->ClusterOffset = 0;
	hFile->Flags &= ~FILE_HANDLE_FLAG_EOF;
}
	
#if FAT_FILE_HELPER
	
	void Fat_Date(fat_date_t* date, uint16_t year, uint8_t month, uint8_t day) {
		*date = (year - 1980) & 0x7F;
		*date <<= 4;
		*date += month & 0x0F;
		*date <<= 5;
		*date += day & 0x1F;	
	}
	void Fat_Time(fat_time_t* time, uint8_t hours, uint8_t minutes, uint8_t seconds) {
		*time = hours & 0x1F;
		*time <<= 6;
		*time += minutes & 0x3F;
		*time <<= 5;
		*time += (seconds >> 1) & 0x1F;
	}
	#if FAT_FIND_SIMPLE
		bool Fat_FileExists(
			const fat_state_t* pFS,
			file_handle_t* hParent,
			const char* fileName
		) {
			char			fatName[11];
			struct fat_direntry	dirEntry;
			
			Fat_CanonicalNameToFAT(fatName, fileName);
			File_Reset(hParent);
			return Fat_FindFileSimple(pFS, hParent, fatName, 0xFF, &dirEntry, NULL);
		}	
	#endif
#endif

#if FAT_FILE_HELPER || FAT_FIND_SIMPLE
	void Fat_CanonicalNameToFAT(char* fatName, register const char* fileName) {
		register int	j;
		register char*	f = fatName;
		memset(fatName, 0x20, 11);
		for( j = 0; j < 11; fileName++, f++, j++ ) {
			if( *fileName == 0 )
				return;
			else if( *fileName == '.' )
				f = fatName + 7;
			else if( *fileName > 0x20 )
				*f = *fileName;
		}
	}	
#endif

uint32_t _Fat_GetFilePos(const fat_state_t* pFS, const file_handle_t* hFile) {
	uint32_t address;

	if( FILE_HANDLE_INVALID(hFile) )
		return 0; 

	//
	// WORD sector adress calc
	//
	if( hFile->StartCA == 1 ) {
		// FAT16 Root Dir (comes before "first cluster")
		// for FAT32 RootDirSz is always 0
		address = pFS->FirstClusterSA - pFS->RootDirSz;		// absolute SA patch for fat16 root dir
		address <<= pFS->BytsPerSec_bit;
	} else {
		// cluster to relative sector address
		_Fat_getClusterPos(pFS, &hFile->CurrentCA, &address);
	}
	//
	// DWORD byte address
	// 
	address += hFile->ClusterOffset;	// add offset (+ allows for more than one cluster offset)
	
	return address;
}

void File_Seek(
	const fat_state_t* pFS,
	file_handle_t* hFile,
	const uint32_t count,
	unsigned long *pAddress
) {
	uint32_t	offset;
	uint16_t	bpc = 1 << pFS->BytsPerClus_bit;
	
	//
	// Check for EOF
	//
	hFile->Position += count;
	if( hFile->Flags & FILE_HANDLE_FLAG_DIRECTORY ) {
		if( hFile->StartCA == 1 ) {
			// FAT12/16 root directory
			hFile->ClusterOffset += count;
			if( hFile->Position > hFile->Size )
				goto _FatIncFilePos_EOF;
			else {
				if(pAddress)
					*pAddress += count;
				return;
			}
		}
	} else {
		// Normal file
		if( hFile->Position > hFile->Size ) {
			goto _FatIncFilePos_EOF;
		}
	}
	
	//
	// Medium position
	//
	offset = count + hFile->ClusterOffset;
	if( offset >= bpc ) {
		fat_ca_t nextCA;
		do {	
			if( _fat_getNextCA( pFS, hFile->CurrentCA, &nextCA ) ) {
				hFile->CurrentCA = nextCA;
				offset -= bpc;
			} else {
				goto _FatIncFilePos_EOF;
			}
		} while( offset >= bpc );

		hFile->ClusterOffset = *((uint16_t*)&offset);
		if( pAddress )
			*pAddress = _Fat_GetFilePos(pFS, hFile);

	} else {
		hFile->ClusterOffset = *((uint16_t*)&offset);
		if( pAddress )
			*pAddress += count;
	}
	return;
	
 _FatIncFilePos_EOF:
	hFile->Flags |= FILE_HANDLE_FLAG_EOF;
}


///@}
