/*
Copyright 2007, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2007
*/

/**
 * @file	ScatterWeb.SHT11.c
 * @ingroup	libSHT11
 * @brief	SHT11 library
 * 
 * @author	Bastian Blywis	<blywis@inf.fu-berlin.de>
 * @date	Aug 2007
 * @since	1.1
 * @version	$Revision$
 * 
 * Initialisation and basic functions for read and write access
 */

#include "ScatterWeb.SHT11.internals.h"
#include "ScatterWeb.SHT11.h"

/******************************************************************************
 * @name	Initialization and configuration
 * @{
 */

void SHT11_setResolution(SHT11_resolution res) {
	bool error = false;
	uint8_t reg;
	error |= SHT11_read_statusreg(&reg);
	if(error) {
		LOG_HIGH("could not read SHT11 status register");
		return;
	}

	if(res == SHT11_high)
		reg &= 0xFE;
	else
		reg |= 0x01;
	
	error |= SHT11_write_statusreg(reg);

	SHT11_data.resolution = res;
}

void libSHT11_Init() {
#warning SHT11 library still needs work regarding timing and wait cycles
	P3DIR |= 0x20;		// set CLK line as output
	SHT11_connectionreset();
	if(SHT11_checkForSHT11())
		SHT11_data.sht11 = true;
	else
		SHT11_data.sht11 = false;
	SHT11_data.crc = 0x00;
	SHT11_data.resolution = SHT11_high;
	SHT11_data.busy = false;
}

bool SHT11_checkForSHT11() {
#warning SHT11_checkForSHT11() not yet implemented (always returns true)
	return true;
}

//@}

bool SHT11_write_byte(uint8_t value) {
	uint8_t i;
	bool error=false;

	SHT11_updateCRC8(value);

	for(i=0; i<8; i++){
		if(value & BIT7)
			SHT11_DATA_HIGH();
		else
			SHT11_DATA_LOW();

		SHT11_CLK_HIGH();		// clk for SENSI-BUS
		_NOP();
		_NOP();
		_NOP();							// pulswith approx. 5 us
		SHT11_CLK_LOW();
		value <<= 1;
	}
	SHT11_CLK_HIGH();
	_NOP();
	_NOP();
	_NOP();								// pulswith approx. 5 us
	SHT11_INPUT();					// SHT11 control data line now
	_NOP();								// clk #9 for ack

	error = SHT11_DATA();		// check ack (DATA will be pulled down by SHT11 on propper command reception)
	SHT11_CLK_LOW();			// data returns to high and control back to uC

	if(error == true)
		LOG_MEDIUM("error writing to SHT11 Sensibus");
	else
		LOG_HIGH("byte successfully written to SHT11 Sensibus");

	return error;
}

uint8_t SHT11_read_byte(bool crcUpdate) {
	SHT11_INPUT();
	uint8_t i,value=0;

	// read data
	for(i=0; i<8; i++) {
		value = value << 1;
		SHT11_CLK_HIGH();
		if(SHT11_DATA())
			value++;
		SHT11_CLK_LOW();
	}
	// data read

	SHT11_OUTPUT();
	_NOP();
	SHT11_DATA_LOW();			// send ack

	SHT11_CLK_HIGH();			//clk #9 for ack
	_NOP();
	_NOP();
	_NOP();								//pulswith approx. 5 us
	SHT11_CLK_LOW();

	if(crcUpdate == true)
		SHT11_updateCRC8(value);

	return value;
}

void SHT11_transstart() {
	SHT11_OUTPUT();

	SHT11_DATA_HIGH();
	SHT11_CLK_LOW();
	_NOP();
	SHT11_CLK_HIGH();
	_NOP();
	SHT11_DATA_LOW();
	_NOP();
	SHT11_CLK_LOW();
	_NOP();
	_NOP();
	_NOP();
	SHT11_CLK_HIGH();
	_NOP();
	SHT11_DATA_HIGH();
	_NOP();
	SHT11_CLK_LOW();
}

void SHT11_connectionreset() {
	SHT11_DATA_HIGH();
	SHT11_CLK_LOW();

	uint8_t i;
	for(i=0;i<9;i++){				//9 SCK cycles
		SHT11_CLK_HIGH();
		SHT11_CLK_LOW();
	}

	SHT11_transstart();
}

bool SHT11_softreset() {
	bool error = false;
	error |= SHT11_write_byte(RESET_SHT11);
	System_delay(11);			// wait 11ms
	return error;
}

bool SHT11_read_statusreg(uint8_t*p_value) {
	bool error=false;
	SHT11_transstart();												//transmission start
	error = SHT11_write_byte(STATUS_REG_R);			//send command to sensor
	*p_value = SHT11_read_byte(true);						//read status register (8-bit)
	uint8_t checksum = SHT11_read_byte(false);		//read checksum (8-bit)

	error |= SHT11_checkCRC(checksum);

	return error;														//error=1 in case of no response form the sensor
}

bool SHT11_write_statusreg(uint8_t value) {
	bool error=false;
	SHT11_transstart();
	error |=SHT11_write_byte(STATUS_REG_W);	//send command to sensor
	error |=SHT11_write_byte(value);						//send value of status register
	return error;
}

bool SHT11_measure(uint16_t* value, SHT11_command command) {
	bool error=false;

	SHT11_transstart();
	switch(command) {
		case MEASURE_TEMP:
			error = SHT11_write_byte(MEASURE_TEMP);
			break;
		case MEASURE_HUMI:
			error = SHT11_write_byte(MEASURE_HUMI);
			break;
		default:
			break;
	}

	/* from the manual: "...takes approximately 11/55/210 ms for
	a 8/12/14bit measurement. The exact time varies by up to 15%
	with the speed of the internal oscillator." */
	while(SHT11_DATA())
		nop();	// TODO yield during conversion

	*value = SHT11_read_byte(true) << 8;				//read the first byte (MSB)
	*value  |= SHT11_read_byte(true);					//read the second byte (LSB)
	uint8_t checksum = SHT11_read_byte(false);	//read checksum

	error |= SHT11_checkCRC(checksum);

	if(error == true)
		LOG_MEDIUM("measurement error");
	return error;
}

float SHT11_temperature(SHT11_tempUnit unit) {
	if(SHT11_data.sht11 == false)
		return SHT11_missing;

	if(SHT11_data.busy == true)
		return SHT11_busy;

	SHT11_data.busy = true;
	uint16_t rawTemp;
	if(SHT11_measure(&rawTemp, MEASURE_TEMP))
		return SHT11_error;

	SHT11_data.busy = false;
	return SHT11_calcTemp(rawTemp, unit);
};

float SHT11_humidityUncorrected() {
	if(SHT11_data.sht11 == false)
		return SHT11_missing;

	if(SHT11_data.busy == true)
		return SHT11_busy;

	SHT11_data.busy = true;
	uint16_t rawHumi;
	if(SHT11_measure(&rawHumi, MEASURE_HUMI))
		return SHT11_error;

	SHT11_data.busy = false;
	return SHT11_calcHumi(rawHumi);
};

float SHT11_humidity() {
	if(SHT11_data.sht11 == false)
		return SHT11_missing;

	if(SHT11_data.busy == true)
		return SHT11_busy;

	SHT11_data.busy = true;
	uint16_t rawHumi;
	uint16_t rawTemp;
	if(SHT11_measure(&rawHumi, MEASURE_HUMI)
	&& SHT11_measure(&rawTemp, MEASURE_TEMP))
		return SHT11_error;

	float h =  SHT11_calcHumi(rawHumi);
	float t = SHT11_calcTemp(rawTemp, SHT11_celsius);

	SHT11_data.busy = false;
	return SHT11_calcHumiCorrected(rawHumi, h, t);
};

float SHT11_calcTemp(uint16_t temperature, SHT11_tempUnit unit) {
	float D1, D2;
	if(unit == SHT11_celsius)
		D1 = -39.60;					// vdd 3V [C]
	else
		D1 = -39.28;	// vdd 3V [C]

	if(SHT11_data.resolution == SHT11_high) 
		D2=+0.01;					//for 14 Bit @ 3V [C]
	else
		D2=+0.04;					//for 12 Bit @ 3V [C]
	
	LOG_VERBOSE("# SHT11 Temp ticks=%u", temperature);
	
	//calc. temperature from ticks
	return D1 + D2 * (temperature+Configuration.sht11_tempdiff);
}

float SHT11_calcHumi(uint16_t humidity) {
	const float C1=-4.0;					//for 8 & 12 Bit
	if(SHT11_data.resolution == SHT11_high) {
		const float C2=+0.0405;		//for 12 Bit
		const float C3=-0.0000028;	//for 12 Bit
		return C3 * humidity * humidity + C2 * humidity + C1;
	}
	else {
		const float C2=+0.648;			//for 8 Bit
		const float C3=-0.00072;		//for 8 Bit
		return C3 * humidity * humidity + C2 * humidity + C1;
	}
}

float SHT11_calcHumiCorrected(uint16_t rawH, float h, float t) {
	const float T1=+0.01;
	if(SHT11_data.resolution == SHT11_high) {
		const float T2=+0.00008;		//for 12 Bit @ 3V
		return (t-25)*(T1+T2*rawH)+h;
	}
 	else {
		const float T2=+0.00128;		//for 8 Bit @ 3V
		return (t-25)*(T1+T2*rawH)+h;
	}
}

/*float calc_dewpoint(float h,float t) {
	float k, dew_point;
	k = (log10(h)-2)/0.4343 + (17.62*t)/(243.12+t);
	dew_point = 243.12*k/(17.62-k);
	return dew_point;
}*/


bool SHT11_checkCRC(uint8_t c) {
	uint8_t reversed = 0x00;
	int8_t i = 0;
	while(true) {
		if(c & 0x01)
			reversed |= 0x01;
		if(i++ == 7)
			break;
		reversed <<= 1;
		c >>= 1;
	}

	if(reversed == SHT11_data.crc) {
		LOG_HIGH("crc match");
		SHT11_updateCRC8(reversed);
		return false;
	}

	LOG_LOW("crc mismatch -> SHT11 reset");
	SHT11_connectionreset();
	return true;
}

uint8_t * myitoa(uint8_t * buf, uint16_t val, uint8_t pad) {
	int8_t i;
	uint8_t tmp_buf[6];

	for(i=0; val>0; val/=10)
		tmp_buf[i++] = (val % 10) + '0';
	if(i==0)
		tmp_buf[i++] = '0';

	while(i<pad)
		tmp_buf[i++] = ' ';

	while(i>0)
		*buf++ = tmp_buf[--i];

	*buf = 0;

	return buf;
}

void ftoa(uint8_t* buf, double f, uint8_t pad) {
	uint8_t *p;

	f = (f*1000.0 + 0.5)/1000.0;
	p = myitoa(buf,f,pad);
	*p++ = '.';

	f -= (uint16_t) f;
	f = f * 1000.0;
	uint16_t t = f;
	while(t < 100) {
		*p++ = '0';
		t*=10;
		if(!t) {
			*p++ = 0x00;
			return;
		}
	}
	myitoa(p,f,pad);
}

/**
 * @brief Measure current temperature:  <code>temperature</code>
 * @ingroup commands
 * @return temperature in [C]
 * @see ScatterWeb.SHT11.h
 *
 * <p>Measures current temperature via the SHT11 if present on the MSB-430.</p>
 */
COMMAND(temperature, 0, cmdargs ) {
	float f = SHT11_temperature(SHT11_celsius);
	uint8_t buf[10];
	ftoa(buf, f, 0);
	snprintf(cmdargs->response_args, cmdargs->response_bufsize,
			"%sC", buf);
}

/**
 * @brief	Sets temperature delta in ticks for fine calibration
 * @ingroup	commands
 * @return	current value
 */
COMMAND(tempdiff, 0, cmdargs ) {
	if( cmdargs->arg_size > 0 ) {
		Configuration.sht11_tempdiff = (int16_t)strtol(cmdargs->args, NULL, 10);
		Configuration_save();
	}
	
	snprintf(cmdargs->response_args, cmdargs->response_bufsize,
			"%i", Configuration.sht11_tempdiff);
}

/**
 * @brief Measure current humidity:  <code>humidity</code>
 * @ingroup commands
 * @return humidity in [%]
 * @see ScatterWeb.SHT11.h
 *
 * <p>Measures current humidity via the SHT11 if present on the MSB-430.</p>
 */
COMMAND(humidity, 0, cmdargs ) {
	float f = SHT11_humidity();
	uint8_t buf[10];
	ftoa(buf, f, 0);
	
	snprintf(cmdargs->response_args, cmdargs->response_bufsize,
			"%s%%RH", buf);
}

void SHT11_updateCRC8(uint8_t b) {
	uint8_t index = SHT11_data.crc ^= b;
	SHT11_data.crc = crc8_table[index];
}
